/*
    Copyright 2007 Robert Knight <robertknight@gmail.com>
    Copyright 2008 Sebastian Sauer <mail@dipe.org>
    Copyright 2008 Helio Chissini de Castro <helio@kde.org>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301, USA.
*/

// Own
#include "mandrivaapplet/mandrivaapplet.h"
#include "mandrivaapplet/menuview.h"
#include "mandrivaapplet/toptile.h"

// Qt
#include <QLabel>
#include <QComboBox>
#include <QGridLayout>
#include <QGraphicsView>
#include <QMetaObject>
#include <QMetaEnum>
#include <QPointer>
#include <QGraphicsLinearLayout>
#include <QDBusConnection>
#include <QDBusMessage>
#include <QSvgRenderer>
#include <QWidgetAction>
#include <QFile>

// KDE
#include <KIcon>
#include <KConfigDialog>
#include <KMenu>
#include <KProcess>
#include <KStandardDirs>

// Plasma
#include <plasma/widgets/icon.h>
#include <plasma/containment.h>

// Local
#include "core/itemhandlers.h"
#include "core/models.h"
#include "core/applicationmodel.h"
#include "core/favoritesmodel.h"
#include "core/systemmodel.h"
#include "core/recentlyusedmodel.h"
#include "core/leavemodel.h"
#include "core/urlitemlauncher.h"


/// @internal d-pointer class
class MenuLauncherApplet::Private
{
public:
        QPointer<Kickoff::MenuView> menuview;
        Plasma::Icon *icon;
        QPointer<Kickoff::UrlItemLauncher> launcher;

        MenuLauncherApplet::ViewType viewtype;
        MenuLauncherApplet::FormatType formattype;

        QComboBox *viewComboBox, *formatComboBox;

        QList<QAction*> actions;
        QAction* switcher;

        class EventSniffer;
        EventSniffer *eventSniffer;

        Private();

        ~Private() { delete menuview; }

        void addItem(QComboBox* combo, const QString& caption, int index, const QString& icon = QString()) {
            if( icon.isEmpty() ) {
                combo->addItem(caption, index);
            }
            else {
                combo->addItem(KIcon(icon), caption, index);
            }
        }

        void setCurrentItem(QComboBox* combo, int currentIndex) {
            for(int i = combo->count() - 1; i >= 0; --i) {
                if( combo->itemData(i).toInt() == currentIndex ) {
                    combo->setCurrentIndex(i);
                    return;
                }
            }
            if( combo->count() > 0 ) {
                combo->setCurrentIndex(0);
            }
        }

        Kickoff::MenuView *createMenuView(QAbstractItemModel *model = 0) {
            Kickoff::MenuView *view = new Kickoff::MenuView(menuview);
            view->setFormatType( (Kickoff::MenuView::FormatType) formattype );
            if( model ) {
                view->setModel(model);
            }
            return view;
        }

        void addMenu(Kickoff::MenuView *view, bool mergeFirstLevel) {
            QList<QAction*> actions = view->actions();
            foreach(QAction *action, actions) {
                if( action->menu() && mergeFirstLevel ) {
                    QMetaObject::invokeMethod(action->menu(),"aboutToShow"); //fetch the children
                    if( actions.count() > 1 && action->menu()->actions().count() > 0 ) {
                        menuview->addTitle(action->text());
                    }
                    foreach(QAction *a, action->menu()->actions()) {
                        a->setVisible(a->menu() || ! view->indexForAction(a).data(Kickoff::UrlRole).isNull());
                        menuview->addAction(a);
                    }
                }
                else {
                    action->setVisible(action->menu() || ! view->indexForAction(action).data(Kickoff::UrlRole).isNull());
                    menuview->addAction(action);
                }
            }

            // if the model asks us for a reset we can't do much except to invalidate our
            // menuview to be able to rebuild it what is needed to prevent dealing with
            // invalid items.
            // the problem here is, that if the menu is currently displayed, it will just
            // close itself what is evil++ but still better then crashes. anyway, the
            // right(TM) solution would be to introduce logic to update the content of the
            // menu even on a reset.
            connect(view->model(), SIGNAL(modelReset()), menuview, SLOT(deleteLater()));
        }

};

class MenuLauncherApplet::Private::EventSniffer
    : public QObject
{
    public:
        EventSniffer(QObject *parent = 0)
            : QObject(parent) { }
        ~EventSniffer() { }

        bool eventFilter(QObject *object, QEvent *event)
        {
            Q_UNUSED(object);
            if (event->type() == QEvent::Paint ||
                    event->type() == QEvent::KeyPress ||
                    event->type() == QEvent::KeyRelease) { return false; }
            event->accept();
            return true;
        }
};

MenuLauncherApplet::Private::Private() :
    menuview(0),
    launcher(0), 
    switcher(0), 
    eventSniffer(new EventSniffer) {}

MenuLauncherApplet::MenuLauncherApplet(QObject *parent, const QVariantList &args)
    : Plasma::Applet(parent,args),
      d(new Private)
{
    KGlobal::locale()->insertCatalog("plasma_applet_launcher");

    setHasConfigurationInterface(true);
    setBackgroundHints(NoBackground);

    

    d->icon = new Plasma::Icon(true, this);
    d->icon->setFlag(ItemIsMovable, false);
    connect(d->icon, SIGNAL(pressed(bool)), this, SLOT(toggleMenu(bool)));
    connect(this, SIGNAL(activate()), this, SLOT(toggleMenu()));

    d->viewtype = Combined;
    d->formattype = NameDescription;
}

MenuLauncherApplet::~MenuLauncherApplet()
{
    delete d;
}

void MenuLauncherApplet::init()
{
    QGraphicsLinearLayout *layout = new QGraphicsLinearLayout(this);
    layout->setContentsMargins(0, 0, 0, 0);
    layout->setSpacing(0);

    KConfigGroup cg = config();

    {
        QMetaEnum e = metaObject()->enumerator(metaObject()->indexOfEnumerator("ViewType"));
        QByteArray ba = cg.readEntry("view", QByteArray(e.valueToKey(d->viewtype)));
        d->viewtype = (MenuLauncherApplet::ViewType) e.keyToValue(ba);
    }
    {
        QMetaEnum e = metaObject()->enumerator(metaObject()->indexOfEnumerator("FormatType"));
        QByteArray ba = cg.readEntry("format", QByteArray(e.valueToKey(d->formattype)));
        d->formattype = (MenuLauncherApplet::FormatType) e.keyToValue(ba);
    }

    KConfig *cfg = new KConfig("mandrivarc");
    KConfigGroup grp = cfg->group("menu");
    QString fileName = grp.exists() ? 
        grp.readEntry( "menubutton", KStandardDirs::locate("data", "mandriva/pics/mdvbutton.svg") ) :
        KStandardDirs::locate("data", "mandriva/pics/mdvbutton.svg");

    bool haveSvg = false;
    if(QFile::exists(fileName))
    {
        QSvgRenderer *renderer = new QSvgRenderer(fileName);
        if( renderer->isValid() )
        {
            setMinimumWidth( renderer->defaultSize().width() );
            d->icon->setSvg( fileName );
            haveSvg = true;
        }
    }

    if( ! haveSvg )
        d->icon->setIcon(KIcon("start-here-kde"));

    layout->addItem( d->icon );

    Kickoff::UrlItemLauncher::addGlobalHandler(Kickoff::UrlItemLauncher::ExtensionHandler,"desktop",new Kickoff::ServiceItemHandler);
    Kickoff::UrlItemLauncher::addGlobalHandler(Kickoff::UrlItemLauncher::ProtocolHandler, "leave", new Kickoff::LeaveItemHandler);

    if (KService::serviceByStorageId("kde4-kmenuedit.desktop")) {
        QAction* menueditor = new QAction(i18n("Menu Editor"), this);
        d->actions.append(menueditor);
        connect(menueditor, SIGNAL(triggered(bool)), this, SLOT(startMenuEditor()));
    }

    Q_ASSERT( ! d->switcher );
    d->switcher = new QAction(i18n("Switch to Kickoff Menu Style"), this);
    d->actions.append(d->switcher);
    connect(d->switcher, SIGNAL(triggered(bool)), this, SLOT(switchMenuStyle()));

    constraintsEvent(Plasma::ImmutableConstraint);
}

void MenuLauncherApplet::constraintsEvent(Plasma::Constraints constraints)
{
    setBackgroundHints(NoBackground);
    if (constraints & Plasma::FormFactorConstraint) {
        if (formFactor() == Plasma::Planar ||
            formFactor() == Plasma::MediaCenter) {
            //FIXME set correct minimum size
            //setMinimumContentSize(d->icon->sizeFromIconSize(IconSize(KIconLoader::Desktop)));
        } else {
            //setMinimumContentSize(d->icon->sizeFromIconSize(IconSize(KIconLoader::Small)));
        }
    }

    if ((constraints & Plasma::ImmutableConstraint) && d->switcher) {
        d->switcher->setVisible(immutability() == Plasma::Mutable);
    }
}

void MenuLauncherApplet::switchMenuStyle()
{
    if (containment()) {
        containment()->addApplet("launcher", QVariantList(), geometry());
        destroy();
    }
}

void MenuLauncherApplet::startMenuEditor()
{
    KProcess::execute("kmenuedit");
}

void MenuLauncherApplet::runCommand()
{
    QDBusMessage msg = QDBusMessage::createMethodCall("org.kde.krunner", "/App", QString(), "newInstance");
    QDBusConnection::sessionBus().send(msg);
}


void MenuLauncherApplet::createConfigurationInterface(KConfigDialog *parent)
{
    QWidget *p = new QWidget;
    QGridLayout *l = new QGridLayout(p);
    p->setLayout(l);

    QLabel *formatLabel = new QLabel(i18nc("@label:listbox How to present applications in a KMenu-like menu", "Format:"), p);
    l->addWidget(formatLabel, 0, 0);
    d->formatComboBox = new QComboBox(p);
    formatLabel->setBuddy(d->formatComboBox);
    d->addItem(d->formatComboBox, i18nc("@item:inlistbox Format:", "Name Only"), MenuLauncherApplet::Name);
    d->addItem(d->formatComboBox, i18nc("@item:inlistbox Format:", "Description Only"), MenuLauncherApplet::Description);
    d->addItem(d->formatComboBox, i18nc("@item:inlistbox Format:", "Name Description"), MenuLauncherApplet::NameDescription);
    d->addItem(d->formatComboBox, i18nc("@item:inlistbox Format:", "Description (Name)"), MenuLauncherApplet::DescriptionName);
    d->addItem(d->formatComboBox, i18nc("@item:inlistbox Format:", "Name - Description"), MenuLauncherApplet::NameDashDescription);
    l->addWidget(d->formatComboBox, 0, 1);

    l->setColumnStretch(1,1);

    d->setCurrentItem(d->formatComboBox, d->formattype);

    parent->setButtons(KDialog::Ok | KDialog::Cancel | KDialog::Apply);
    connect(parent, SIGNAL(applyClicked()), this, SLOT(configAccepted()));
    connect(parent, SIGNAL(okClicked()), this, SLOT(configAccepted()));
    parent->addPage(p, parent->windowTitle(), icon());
}

void MenuLauncherApplet::configAccepted()
{
    bool needssaving = false;
    KConfigGroup cg = config();

    int ft = d->formatComboBox->itemData(d->formatComboBox->currentIndex()).toInt();
    if( ft != d->formattype ) {
        d->formattype = (MenuLauncherApplet::FormatType) ft;
        needssaving = true;

        QMetaEnum e = metaObject()->enumerator(metaObject()->indexOfEnumerator("FormatType"));
        cg.writeEntry("format", QByteArray(e.valueToKey(d->formattype)));
    }

    if( needssaving ) {
        emit configNeedsSaving();

        delete d->menuview;
        d->menuview = 0;
    }
}

void MenuLauncherApplet::toggleMenu(bool pressed)
{
    if (pressed) {
        toggleMenu();
    }
}

void MenuLauncherApplet::toggleMenu()
{
    if (!d->menuview) {
        d->menuview = new Kickoff::MenuView();
        connect(d->menuview, SIGNAL(triggered(QAction*)), this, SLOT(actionTriggered(QAction*)));
        connect(d->menuview, SIGNAL(aboutToHide()), d->icon, SLOT(setUnpressed()));
        //connect(d->menuview, SIGNAL(aboutToHide()), d->menuview, SLOT(deleteLater()));
        
        Kickoff::ApplicationModel *appModel = new Kickoff::ApplicationModel(d->menuview);
        appModel->setDuplicatePolicy(Kickoff::ApplicationModel::ShowLatestOnlyPolicy);

        //Top view
        KConfig *cfg = new KConfig("mandrivarc");
        KConfigGroup grp = cfg->group("menu");
        QSvgRenderer *renderer;
        QString fileName = grp.exists() ?
            grp.readEntry( "topimage", KStandardDirs::locate("data", "mandriva/pics/top.svg") ) :
            KStandardDirs::locate("data", "mandriva/pics/top.svg");
        if(QFile::exists(fileName))
        {
            renderer = new QSvgRenderer(fileName);
            if( renderer->isValid() )
            {
                QSize t = renderer->defaultSize();
                QPixmap pix(t);
                pix.fill(Qt::transparent);
                QPainter p(&pix);
                renderer->render(&p);
                p.end();
                QWidgetAction *topAction = new QWidgetAction(d->menuview);
                TopTile *top = new TopTile(pix);
                top->installEventFilter(d->eventSniffer);
                topAction->setDefaultWidget( top );
                d->menuview->addAction( topAction );
            }
        }

        d->menuview->addTitle(i18n("All Applications"));
        Kickoff::MenuView *appview = d->createMenuView(appModel);
        d->addMenu(appview, false);

        d->menuview->addTitle( i18n("Recently Used"));
        Kickoff::MenuView *recentlyview = d->createMenuView(new Kickoff::RecentlyUsedModel(d->menuview));
        d->addMenu(recentlyview, false);

        d->menuview->addSeparator();
        d->menuview->addTitle(i18n("Actions"));
        QAction *runaction =  d->menuview->addAction(KIcon("system-run"),i18n("Run Command..."));
        runaction->setData(KUrl("run:/command"));

        d->menuview->addSeparator();
        QAction *switchaction = d->menuview->addAction(KIcon("system-switch-user"),i18n("Switch User"));
        switchaction->setData(KUrl("leave:/switch"));
        QAction *lockaction = d->menuview->addAction(KIcon("system-lock-screen"),i18n("Lock"));
        lockaction->setData(KUrl("leave:/lock"));
        QAction *logoutaction = d->menuview->addAction(KIcon("system-shutdown"),i18n("Leave"));
        logoutaction->setData(KUrl("leave:/logout"));
        if( cfg )
            delete cfg;
    }

    d->menuview->setAttribute(Qt::WA_DeleteOnClose);
    d->menuview->popup(popupPosition(d->menuview->sizeHint()));
    d->icon->setPressed();
}

void MenuLauncherApplet::actionTriggered(QAction *action)
{
    if( action->data().isNull() )
        return;

    KUrl url = action->data().value<KUrl>();
    if (url.scheme() == "leave") {
        if ( ! d->launcher ) {
            d->launcher = new Kickoff::UrlItemLauncher(d->menuview);
        }
        d->launcher->openUrl(url.url());
        return;
    }
    if (url.scheme() == "run")
    {
        runCommand();
        return;
    }
    for(QWidget* w = action->parentWidget(); w; w = w->parentWidget()) {
        if (Kickoff::MenuView *view = dynamic_cast<Kickoff::MenuView*>(w)) {
            view->actionTriggered(action);
            break;
        }
    }
}

QList<QAction*> MenuLauncherApplet::contextualActions()
{
    return d->actions;
}

#include "mandrivaapplet.moc"
