/*
 *   Copyright (C) 2008 Dmitry Suzdalev <dimsuz@gmail.com>
 *
 * This program is free software you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */
#include "mainwidget.h"

#include <plasma/widgets/label.h>
#include <plasma/panelsvg.h>

#include <QPainter>
#include <QBitmap>
#include <QLabel>
#include <QGraphicsLinearLayout>
#include <QStyleOptionGraphicsItem>
#include <QGraphicsSceneResizeEvent>
#include <QDBusMessage>
#include <QDBusConnection>

#include "notificationwidget.h"

MainWidget::MainWidget(QGraphicsItem* parent)
    : Plasma::Containment(parent),
    m_layout(0),
    m_introLabel(0)
{
    m_panelSvg = new Plasma::PanelSvg(this);
    m_panelSvg->setImagePath("dialogs/background");
    m_panelSvg->setEnabledBorders(Plasma::PanelSvg::AllBorders);

    m_layout = new QGraphicsLinearLayout( Qt::Vertical );
    m_layout->setContentsMargins( 0, 0, 0, 0 );
    m_layout->setSpacing(0);

    m_introLabel = new Plasma::Label(this);
    m_introLabel->nativeWidget()->setAlignment(Qt::AlignCenter);
    m_introLabel->nativeWidget()->setText(i18n("No new notifications"));
    m_layout->addItem(m_introLabel);
    // NOTE: due to a bug (or feature) of QGraphicsLinearLayout, it sets QGraphicsWidget's
    // size to (1,1) if layout was added being empty. Here we are safe as we already
    // have a widget in layout. But whoever reads this - be aware ;)
    setLayout(m_layout);
}

MainWidget::~MainWidget()
{
}

void MainWidget::paint ( QPainter * painter, const QStyleOptionGraphicsItem * option, QWidget * widget )
{
    Q_UNUSED(widget)

    // draw vertical lines between widgets, but assure these lines
    // won't reach the very top
    if (m_notifyWidgets.isEmpty()) {
        m_panelSvg->paintPanel(painter, option->rect);
    } else if (m_notifyWidgets.count() > 1) {
        painter->setRenderHint(QPainter::Antialiasing, true);
        QPen pen(Qt::white);
        pen.setWidth(1);
        painter->setPen( pen );

        // NOTE: lines colors and other painter settings used here
        // should be in sync with ones used in NotificationWidget::paint()
        // make them common explicitly somehow?
        const int startY = 10;
        painter->fillRect(0, startY, option->rect.width(), option->rect.height()-startY, Qt::black);
        painter->drawLine(1, startY, 1, option->rect.bottom());
        painter->drawLine(option->rect.width()-1, startY, option->rect.width()-1, option->rect.bottom());
    }
}

void MainWidget::handleNewData( const Plasma::DataEngine::Data &data )
{
    QString appName = data.value( "appName" ).toString();

    NotificationWidget* widget = 0;
    if (!m_notifyWidgets.contains(appName)) {
        widget = new NotificationWidget(this);
        connect( widget, SIGNAL(actionInvoked(uint,uint)), SLOT(onActionInvoked(uint,uint)));

        m_notifyWidgets.insert( appName, widget );
        // prepend an item
        m_layout->insertItem( 0, widget );

        m_introLabel->hide();

        // and now we need to render different set of borders
        m_panelSvg->setEnabledBorders( Plasma::PanelSvg::TopBorder |
                                       Plasma::PanelSvg::LeftBorder |
                                       Plasma::PanelSvg::RightBorder );

        // resize widget if needed to fit all widgets
        updateWidgetSize();
        emit widgetCountChanged();
    } else {
        widget = m_notifyWidgets.value(appName);
    }

    if (widget) {
        uint id = data.value( "id", 0 ).toUInt();
        if (widget->hasNotification(id)) {
            // TODO update it
            kDebug() << "updating notification data is not implemented yet";
        } else {
            widget->addNotification( data );
        }
    }
}

void MainWidget::updateWidgetSize()
{
    int curHeight = size().height();
    int minimumHeight = NotificationWidget::desiredMinimumHeight()*m_notifyWidgets.count();
    if (curHeight < minimumHeight || curHeight > minimumHeight ) {
        QSizeF newSize = QSizeF(size().width(), minimumHeight);
        resize(newSize);
    }
}

void MainWidget::resizeEvent(QGraphicsSceneResizeEvent* event)
{
    m_panelSvg->resizePanel(event->newSize());
    Plasma::Containment::resizeEvent(event);
}

void MainWidget::removeNotification( uint id )
{
    kDebug() << "removing notification" << id;
    NotificationWidget* widget = findWidgetById(id);
    if (widget) {
        widget->removeNotification(id);

        if (widget->notificationsCount() == 0) {
            // if we are deleting last widget then we show intro (that is being hidden)
            // before doing that
            // (we always keep layout filled - see NOTE about setLayout above)
            // "2" stands for: last widget being removed + intro being hidden
            if (m_layout->count() == 2) {
                m_introLabel->show();
                // NOTE: perhaps this is a bug - Plasma::Label::show() doesnt show nativeWidget,
                // hence the need to call it explicitely
                m_introLabel->nativeWidget()->show();

                // and now we need to render different set of borders
                m_panelSvg->setEnabledBorders(Plasma::PanelSvg::AllBorders);

                // NOTE: for some unknown reason m_introLabel is not being shown until update() is
                // called, though i think it should be shown automatically. Some Qt bug perhaps.
                update();
            }

            // delete empty widget
            m_layout->removeItem(widget);
            m_notifyWidgets.remove(widget->notificationsOwner());
            delete widget;
            widget = 0;

            if (!m_introLabel->isVisible()) {
                // shrink widget if needed
                updateWidgetSize();
            }

            emit widgetCountChanged();
        }
    }
    else {
        kDebug() << "failed to remove: item not found";
    }
}

NotificationWidget* MainWidget::findWidgetById(uint id)
{
    foreach (const QString& appName, m_notifyWidgets.keys()) {
        NotificationWidget* widget = m_notifyWidgets.value(appName);
        if (widget->hasNotification(id)) {
            return widget;
        }
    }
    return 0;
}

int MainWidget::notificationCount() const
{
    return m_notifyWidgets.count();
}

QBitmap MainWidget::widgetMask() const
{
    return m_panelSvg->mask();
}

void MainWidget::onActionInvoked(uint id, uint actionId)
{
    kDebug() << "action" << actionId << "invoked on notification" << id;

    // XXX: This code shouldn't be here! Applet should know nothing about DBus
    // interfaces. All this DBus signal sending work should be done by 
    // Notifications DataEngine because it actually implements this DBus Interface.
    //
    // What Applet should be able to do is to simply tell engine: 
    // "hey, this action has been triggered, do something if you wish"
    //
    // Currently there's no way to do this. aseigo said it would be possible
    // with Plasma::Service. Until it is implemented I put this code here.
    // Consider this a temporary workaround that works for now ;)

    QDBusMessage sig = QDBusMessage::createSignal( "/VisualNotifications", "org.kde.VisualNotifications", "ActionInvoked" );
    QList<QVariant> args;
    args.append( id );
    args.append( QString::number(actionId) ); // actKey param is string

    sig.setArguments( args );

    bool queued = QDBusConnection::sessionBus().send(sig);
    if (!queued) {
        kDebug() << "warning: failed to send DBus signal";
    }
}
