/*
 *   Copyright (C) 2008 Dmitry Suzdalev <dimsuz@gmail.com>
 *
 * This program is free software you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
*/
#include "notificationwidget.h"

#include "notifytextitem.h"

#include <QPainter>
#include <QStyleOptionGraphicsItem>
#include <QGraphicsSceneResizeEvent>
#include <KGlobalSettings>

#include <plasma/theme.h>
#include <plasma/panelsvg.h>
#include <plasma/widgets/icon.h>

static const int ICON_SIZE = 32;
static const int HOR_SPACING = 5;
static const int HEADING_HEIGHT = 24;
static const int GRP_BTN_SIZE = 16;
static const int MID_ARROW_MARGIN = 3;

NotificationWidget::NotificationWidget( QGraphicsItem *parent )
    : QGraphicsWidget( parent ), m_currentIdx(-1), m_oldIdx(-1),
    m_bodyTextX(0), m_counterWidth(0),
    m_leftArrowIcon(0), m_rightArrowIcon(0),
    m_clipItem(0),
    m_animating(false)
{
    // FIXME: as all widgets share same sizes, this should be created only once for all widgets!
    m_panelSvg = new Plasma::PanelSvg(this);
    m_panelSvg->setImagePath("dialogs/background");
    m_panelSvg->setEnabledBorders( Plasma::PanelSvg::TopBorder |
                                   Plasma::PanelSvg::LeftBorder |
                                   Plasma::PanelSvg::RightBorder );

    setMinimumHeight(desiredMinimumHeight());
    connect(Plasma::Animator::self(), SIGNAL(customAnimationFinished(int)), SLOT(onAnimationFinished()));
}

NotificationWidget::~NotificationWidget()
{
    delete m_leftArrowIcon;
    delete m_rightArrowIcon;
}

int NotificationWidget::desiredMinimumHeight()
{
    QFont smallFont = KGlobalSettings::smallestReadableFont();
    QFontMetrics fm(smallFont);

    const int MARGIN = 4;

    return MARGIN*2 + HEADING_HEIGHT + fm.height()*4;
}

void NotificationWidget::addNotification( const Plasma::DataEngine::Data& data )
{
    if ( data.isEmpty() ) {
        kDebug() << "invalid data received";
        return;
    }

    // set common data only once
    if (m_notifyItems.isEmpty()) {
        m_appIcon = KIcon( data.value( "appIcon", "unknown" ).toString() );
        m_appName = data.value( "appName" ).toString();
    }


    uint id = data.value( "id", 0 ).toUInt();
    QString summary = data.value( "summary" ).toString();
    QString body = data.value( "body" ).toString();
    int timeout = data.value( "expireTimeout" ).toInt();
    QStringList actions = data.value( "actions" ).toStringList();

    // create a clip item that will be used to restrict animations to certain rect
    if (!m_clipItem) {
        m_clipItem = new QGraphicsRectItem(this);
        m_clipItem->show();
    }

    // give it a clip item as a parent so it will clip item painting to its rect
    NotifyTextItem* item = new NotifyTextItem(m_clipItem);
    item->hide();
    item->setId(id);
    item->setNotificationText(m_appName, summary, body);
    item->setActions(actions);
    item->setTimeout(timeout);

    // forward signals
    connect(item, SIGNAL(actionInvoked(uint,uint)), SIGNAL(actionInvoked(uint,uint)));

    m_notifyItems.append(item);

    setCurrentItem(m_notifyItems.count()-1);
    updateLayout(size());
}

void NotificationWidget::removeNotification(uint id)
{
    if (m_animating) {
        // will be deleted after animation finishes
        m_itemsToRemove.append(id);
    } else {
        removeNotificationInstantly(id);
    }
}

void NotificationWidget::removeNotificationInstantly(uint id)
{
    // we assume here that we can have only one notification
    // with given id in list
    int idxToRemove = -1;
    for (int i=0; i<m_notifyItems.count(); ++i) {
        NotifyTextItem* item = m_notifyItems.at(i);
        if (item->id() == id) {
            idxToRemove = i;
            break;
        }
    }
    if (idxToRemove != -1) {
        NotifyTextItem* item = m_notifyItems.at(idxToRemove);
        m_notifyItems.removeAt(idxToRemove);
        delete item;
    }
    else {
        kDebug() << "failed to remove notification with id"<< id;
    }

    setCurrentItem(m_notifyItems.count()-1, false);
    updateLayout(size());
}

void NotificationWidget::setCurrentItem(int idx, bool animate)
{
    if (idx == -1) {
        m_currentIdx = -1;
        return;
    }

    m_oldIdx = m_currentIdx;

    m_currentIdx = idx;

    if (animate && m_currentIdx != -1 && m_oldIdx != -1) {
        int frames = m_notifyItems.at(m_currentIdx)->boundingRect().width();
        Plasma::Animator::self()->customAnimation( frames, 300, Plasma::Animator::EaseInCurve, this, "animationFrame" );
        m_animating = true;
    }
    else // when it is needed to change current item w/o animation
    {
        updateAfterSwitchToItem();
    }
}

QString NotificationWidget::notificationsOwner() const
{
    return m_appName;
}

bool NotificationWidget::hasNotification(uint id) const
{
    foreach (NotifyTextItem *item, m_notifyItems) {
        if (item->id() == id) {
            return true;
        }
    }
    return false;
}

int NotificationWidget::notificationsCount() const
{
    return m_notifyItems.count();
}

void NotificationWidget::paint( QPainter * painter, const QStyleOptionGraphicsItem * option, QWidget *)
{
    painter->setRenderHint( QPainter::Antialiasing, true );
    QColor bkgndCol = Plasma::Theme::defaultTheme()->color(Plasma::Theme::BackgroundColor);
    QColor hlCol = Plasma::Theme::defaultTheme()->color(Plasma::Theme::HighlightColor);
    QColor textCol = Plasma::Theme::defaultTheme()->color(Plasma::Theme::TextColor);

    m_panelSvg->paintPanel(painter,option->rect);

    qreal leftMargin, topMargin, rightMargin, bottomMargin;
    m_panelSvg->getMargins(leftMargin, topMargin, rightMargin, bottomMargin);

    /*
     * Draw heading
     */
    QRectF headingRect = QRectF(leftMargin, topMargin, 
                                option->rect.width()-leftMargin-rightMargin,
                                HEADING_HEIGHT);

    QLinearGradient grad(headingRect.topLeft(), headingRect.topRight());
    grad.setColorAt(0, bkgndCol);
    grad.setColorAt(0.3, hlCol);
    grad.setColorAt(0.7, hlCol);
    grad.setColorAt(1, bkgndCol);
    painter->fillRect(headingRect, grad);

    /*
     * Draw icon
     */
    painter->drawPixmap( leftMargin, topMargin + HEADING_HEIGHT + 5, m_appIcon.pixmap( ICON_SIZE ) );

    /*
     * Draw summary text
     */
    painter->setPen( textCol );
    drawSummaryText(painter, headingRect);

    QFont smallFont = KGlobalSettings::smallestReadableFont();

    /*
     * Draw message count marker
     */
    int nCount = m_notifyItems.count();
    if (nCount > 1) {
        QRectF labelRect;
        labelRect.setX(headingRect.right()-GRP_BTN_SIZE - m_counterWidth);
        labelRect.setY(headingRect.y());
        labelRect.setWidth(m_counterWidth);
        labelRect.setHeight(headingRect.height());

        QString labelText = QString("%1 of %2").arg(m_currentIdx+1).arg(nCount);
        painter->drawText(labelRect, Qt::AlignCenter, labelText);
    }
}

void NotificationWidget::drawSummaryText(QPainter* painter, const QRectF& headingRect)
{
    QFont smallFont = KGlobalSettings::smallestReadableFont();
    painter->setFont(smallFont);

    int textFlags = Qt::AlignLeft | Qt::AlignVCenter;
    smallFont.setBold(true);
    painter->setFont(smallFont);

    int nCount = m_notifyItems.count();
    int maxTextWidth = nCount > 1 ? m_leftArrowIcon->pos().x()-headingRect.x() : headingRect.width();

    int curWidth = painter->fontMetrics().width(m_summaryText);
    if (curWidth < maxTextWidth) {
        painter->drawText( headingRect, textFlags, m_summaryText );
    } else {
        QPixmap pixmap(headingRect.size().toSize());
        pixmap.fill(Qt::transparent);

        QPainter p(&pixmap);
        p.setPen(painter->pen());
        p.setFont(painter->font());

        // Create the alpha gradient for the fade out effect
        QLinearGradient alphaGradient(0, 0, 1, 0);
        alphaGradient.setCoordinateMode(QGradient::ObjectBoundingMode);

        alphaGradient.setColorAt(0, QColor(0, 0, 0, 255));
        alphaGradient.setColorAt(1, QColor(0, 0, 0, 0));

        QFontMetrics fm = p.fontMetrics();
        int fadeWidth = 50;

        p.drawText( QRectF(0,0, maxTextWidth, headingRect.height()), textFlags, m_summaryText );

        // Reduce the alpha using the alpha gradient
        p.setCompositionMode(QPainter::CompositionMode_DestinationIn);
        QRectF fadeOutRect( maxTextWidth-fadeWidth, 5, fadeWidth, headingRect.height());
        p.fillRect(fadeOutRect, alphaGradient);

        p.end();

        painter->drawPixmap(headingRect.topLeft(), pixmap);
    }

    smallFont.setBold(false);
    painter->setFont(smallFont);
}

void NotificationWidget::resizeEvent(QGraphicsSceneResizeEvent* event)
{
    m_panelSvg->resizePanel(event->newSize());
    updateLayout(event->newSize());
}

void NotificationWidget::updateLayout( const QSizeF& newSize )
{
    qreal leftMargin, topMargin, rightMargin, bottomMargin;
    m_panelSvg->getMargins(leftMargin, topMargin, rightMargin, bottomMargin);

    // layout: margin, icon, spacing, TEXT, margin
    m_bodyTextX = leftMargin + ICON_SIZE + HOR_SPACING;

    if (m_counterWidth == 0) {
        QFont smallFont = KGlobalSettings::smallestReadableFont();
        QFontMetrics smallFm(smallFont);
        m_counterWidth = smallFm.width( "1 of 100" ) + 2*MID_ARROW_MARGIN;
    }

    if ( m_leftArrowIcon == 0 && m_rightArrowIcon == 0 ) {
        m_leftArrowIcon = new Plasma::Icon(this);
        m_leftArrowIcon->setSvg("notify/left");
        m_leftArrowIcon->resize(GRP_BTN_SIZE, GRP_BTN_SIZE);

        m_rightArrowIcon = new Plasma::Icon(this);
        m_rightArrowIcon->setSvg("notify/right");
        m_rightArrowIcon->resize(GRP_BTN_SIZE, GRP_BTN_SIZE);

        connect(m_leftArrowIcon, SIGNAL(clicked()), SLOT(previousMessage()));
        connect(m_rightArrowIcon, SIGNAL(clicked()), SLOT(nextMessage()));

        kDebug() << "setting pos:" << m_leftArrowIcon->pos();
    }

    m_leftArrowIcon->setPos(newSize.width() - GRP_BTN_SIZE*2 - rightMargin - m_counterWidth, topMargin + HEADING_HEIGHT/2 - GRP_BTN_SIZE/2);
    m_rightArrowIcon->setPos(newSize.width() - GRP_BTN_SIZE - rightMargin, topMargin + HEADING_HEIGHT/2 - GRP_BTN_SIZE/2);

    int nCount = m_notifyItems.count();
    m_leftArrowIcon->setVisible( nCount > 1 );
    m_rightArrowIcon->setVisible( nCount > 1 );

    updateItemGeometries(newSize);

    // repaint now
    update();
}

void NotificationWidget::updateItemGeometries(const QSizeF& size)
{
    qreal leftMargin, topMargin, rightMargin, bottomMargin;
    m_panelSvg->getMargins(leftMargin, topMargin, rightMargin, bottomMargin);

    int bodyRectWidth = size.width() - leftMargin - rightMargin - ICON_SIZE - HOR_SPACING;
    int bodyRectHeight = size.height() - topMargin - bottomMargin - HEADING_HEIGHT;

    // setup a clip item
    if (m_clipItem) {
        m_clipItem->setRect(m_bodyTextX, topMargin+HEADING_HEIGHT, bodyRectWidth, bodyRectHeight);
        m_clipItem->setFlag(QGraphicsItem::ItemClipsChildrenToShape, true);
        m_clipItem->setPen(Qt::NoPen);
    }

    // update item positions
    foreach (NotifyTextItem* item, m_notifyItems) {
        item->setItemSize(bodyRectWidth, bodyRectHeight);
        item->setPos(m_bodyTextX, topMargin + HEADING_HEIGHT);
    }
}

void NotificationWidget::previousMessage()
{
    if (m_currentIdx==0) {
        return;
    }

    setCurrentItem(m_currentIdx-1);
    update();
}

void NotificationWidget::nextMessage()
{
    if (m_currentIdx == m_notifyItems.count()-1) {
        return;
    }

    setCurrentItem(m_currentIdx+1);
    update();
}

void NotificationWidget::animationFrame(qreal progress)
{
    if (m_currentIdx == -1 || m_oldIdx == -1 || !m_clipItem) {
        return;
    }

    NotifyTextItem* curItem = m_notifyItems.at(m_oldIdx);
    if (!curItem) {
        return;
    }

    NotifyTextItem* newItem = m_notifyItems.at(m_currentIdx);
    if (!newItem->isVisible()) {
        newItem->show();
    }
    if (!curItem->isVisible()) {
        curItem->show();
    }

    int width = curItem->boundingRect().width();
    int direction = (m_currentIdx > m_oldIdx) ? -1 : 1;
    curItem->setPos(m_bodyTextX+direction*width*progress, curItem->pos().y());

    if (newItem) {
        newItem->setPos(curItem->pos().x() - direction*width, curItem->pos().y());
    }
}

void NotificationWidget::onAnimationFinished()
{
    m_animating = false;

    // delete items that were put on hold until animatin finishes
    foreach (uint id, m_itemsToRemove) {
        removeNotificationInstantly(id);
    }
    m_itemsToRemove.clear();

    updateAfterSwitchToItem();
}

void NotificationWidget::updateAfterSwitchToItem()
{
    foreach (NotifyTextItem* item, m_notifyItems) {
        item->hide();
    }

    m_notifyItems.at(m_currentIdx)->show();
    m_summaryText = m_notifyItems.at(m_currentIdx)->summary();
    m_oldIdx = -1;
}
