/*
 *   Copyright (C) 2008 Dmitry Suzdalev <dimsuz@gmail.com>
 *
 * This program is free software you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
*/
#ifndef NOTIFICATIONITEM_H
#define NOTIFICATIONITEM_H

#include <plasma/dataengine.h>

#include <QGraphicsWidget>
#include <KIcon>
#include <QList>

namespace Plasma
{
    class Svg;
    class Icon;
    class PanelSvg;
} // namespace Plasma

class QToolButton;
class QGraphicsProxyWidget;
class NotifyTextItem;

/**
 *  A graphics item, representing notification message
 */
class NotificationWidget : public QGraphicsWidget
{
    Q_OBJECT
public:
    NotificationWidget( QGraphicsItem *parent );
    ~NotificationWidget();

    /**
     * Adds notification to this widget.
     * If more than one notification is added to the widget,
     * its UI will be adjusted to allow scrolling through all
     * of added notifications
     */
    void addNotification( const Plasma::DataEngine::Data& data );
    /**
     * Removes notification with specified id from this widget or schedules it for deletion 
     * if some animation is in progress
     * Note: if this is the last notification in this widget, widget won't be deleted,
     * it'll remain empty. Calling code should take care of deleting empty widget
     * if it wants
     */
    void removeNotification(uint id);

    /**
     * Retrieves the name of the application owning the notifications
     * displayed in this widget
     */
    QString notificationsOwner() const;

    /**
     * Returns true if widget contains notification with specified id
     */
    bool hasNotification(uint id) const;

    /**
     * Returns number of notifications displayed in this widget
     */
    int notificationsCount() const;

    virtual void paint( QPainter * painter, const QStyleOptionGraphicsItem * option, QWidget * widget = 0 );

    static int desiredMinimumHeight();

signals:
    void actionInvoked(uint id, uint actionId);

private slots:
    void previousMessage();
    void nextMessage();

    void animationFrame(qreal progress);

    void onAnimationFinished();

private:
    virtual void resizeEvent(QGraphicsSceneResizeEvent* event);

    /**
     * Removes notification from widget at once.
     * @see removeNotification
     */
    void removeNotificationInstantly(uint id);

    void updateAfterSwitchToItem();

    void setCurrentItem(int idx, bool animate = true);
    void updateLayout( const QSizeF& newSize );
    void updateItemGeometries(const QSizeF& size);
    void drawSummaryText(QPainter*, const QRectF&);

private:
    int m_currentIdx;
    // while animating this will hold the index of an old item (which goes away)
    int m_oldIdx;

    KIcon m_appIcon;
    QString m_appName;

    QList<NotifyTextItem*> m_notifyItems;

    int m_bodyTextX;
    int m_counterWidth;

    // these are the properties of currently shown notification
    QString m_summaryText;

    Plasma::Icon* m_leftArrowIcon;
    Plasma::Icon* m_rightArrowIcon;

    Plasma::PanelSvg* m_panelSvg;

    QGraphicsRectItem* m_clipItem;

    /**
     * Holds true if some animation is currently running
     */
    bool m_animating;

    /**
     *  List of the items to be removed after animation finishes
     */
    QList<uint> m_itemsToRemove;
};

#endif // NOTIFICATIONITEM_H
