/*
 *   Copyright (C) 2008 Dmitry Suzdalev <dimsuz@gmail.com>
 *
 * This program is free software you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
*/

#include "notifytextitem.h"

#include <QPainter>
#include <QCursor>
#include <QGraphicsSceneMouseEvent>
#include <QTextDocument>
#include <QFontMetrics>
#include <QAbstractTextDocumentLayout>
#include <KGlobalSettings>
#include <KLocale>
#include <KDebug>

#include <plasma/theme.h>

NotifyTextItem::NotifyTextItem(QGraphicsItem* parent)
    : QGraphicsItem(parent), m_size(30,30)
{
    m_bodyTextDoc = new QTextDocument();
    m_bodyTextDoc->setDefaultFont(KGlobalSettings::smallestReadableFont());
}

NotifyTextItem::~NotifyTextItem()
{
    delete m_bodyTextDoc;
}

QRectF NotifyTextItem::boundingRect() const
{
    return QRectF(QPointF(0,0), m_size);
}

void NotifyTextItem::setItemSize(qreal width, qreal height)
{
    prepareGeometryChange();
    m_size = QSizeF(width, height);
    m_bodyTextDoc->setTextWidth(width);

    updateActionRects();
}

void NotifyTextItem::paint(QPainter* painter, const QStyleOptionGraphicsItem *option, QWidget* widget)
{
    Q_UNUSED(option)
    Q_UNUSED(widget)

    painter->setRenderHint( QPainter::Antialiasing, true );

    QRectF boundRect = boundingRect();

    QColor hlCol = Plasma::Theme::defaultTheme()->color(Plasma::Theme::HighlightColor);
    QColor textCol = Plasma::Theme::defaultTheme()->color(Plasma::Theme::TextColor);

    qreal actHeight = 0;
    if (!m_actions.isEmpty()) {
        QFont smallFont = KGlobalSettings::smallestReadableFont();
        painter->setFont(smallFont);

        actHeight = m_actions.at(0).actRect.height();

        QBrush oldBrush = painter->brush();
        QPen oldPen = painter->pen();
        foreach (const ActionInfo& action, m_actions) {
            painter->setBrush( m_hoveredActId == action.actId ? hlCol.lighter(350) : hlCol.lighter(280) );
            painter->setPen( hlCol.darker() );
            painter->drawRoundedRect(action.actRect, 3, 3);
            painter->drawText(action.actRect, Qt::AlignCenter, action.actName);
        }
        painter->setBrush(oldBrush);
        painter->setPen(oldPen);
    }

    QAbstractTextDocumentLayout::PaintContext paintContext;
    paintContext.palette.setColor( QPalette::Text, textCol );
    paintContext.clip = boundRect.adjusted(0,0,0,-actHeight);

    m_bodyTextDoc->documentLayout()->draw(painter, paintContext);
}

void NotifyTextItem::setNotificationText(const QString& appName, const QString& summary, const QString& body)
{
    m_bodyTextDoc->setHtml(body);

    if ( summary.isEmpty() && !body.isEmpty() ) {
        // Do some magic to extract first line from body text and put
        // it to the summary so it won't be empty.
        // It's not that straigh easy because body text can be either plain text
        // or html text

        // let's find first line in plain version of text
        QString plainText = m_bodyTextDoc->toPlainText();
        int idx = plainText.indexOf('\n');
        if (idx != -1) {
            // okay, text consists of several lines. We want to put first
            // line to summary and hence remove it from body.
            // We'll do the following:
            // - extract first line to put it in summary
            // - get the html representation of text
            // - simply remove the text we extracted and remove first <br/>
            //   (as we assume that first br will follow first line - this
            //    allows us not to mess with all possible combinations of tags)
            // - update document with modified text
            m_summary = plainText.left(idx);
            QString html = m_bodyTextDoc->toHtml();
            html.remove(m_summary);
            QRegExp regexp("<br\\s*/?\\s*>");
            int brIdx = regexp.indexIn(html);
            if (brIdx != -1) {
                html.remove(brIdx, regexp.matchedLength());
            }
            m_bodyTextDoc->setHtml(html);
        } else {
            m_summary = i18n("Notification from ")+appName;
        }
    } else {
        m_summary = summary;
    }

}

void NotifyTextItem::setActions(const QStringList& actions)
{
    if (!actions.isEmpty() ) {
        setAcceptsHoverEvents(true);
        int actCount = actions.count();
        // we expect actions to be list of id,action,id,action,...
        if (actCount % 2 == 0) {
            uint actKey;
            QString actName;
            for (int idx=0; idx<actCount; ++idx) {
                // each even element is action key, each odd is action name
                if (idx % 2 == 0) {
                    actKey = actions.at(idx).toUInt();
                } else {
                    actName = actions.at(idx);
                    m_actions.append(ActionInfo(actKey, actName));
                }
            }
            updateActionRects();
        }
        else {
            kDebug() << "error: wrongly formatted action array was set";
        }
    }
}

void NotifyTextItem::updateActionRects()
{
    QFont smallFont = KGlobalSettings::smallestReadableFont();
    QFontMetrics fm(smallFont);
    int actHeight = fm.height();

    int numAct = m_actions.count();

    QRectF boundRect = boundingRect();
    qreal actX = boundRect.right();
    const int midSpacing = 4;

    for (int idx=numAct-1; idx >=0; --idx) {
        int width = fm.width(m_actions.at(idx).actName)+4;
        actX -= width+midSpacing;
        m_actions[idx].actRect = QRectF(actX, boundRect.bottom()-actHeight-2, width, actHeight);
    }
}

void NotifyTextItem::mousePressEvent(QGraphicsSceneMouseEvent*)
{
    // dummy reimplementation needed to accept mouse event to be able to receive release events
}

void NotifyTextItem::mouseReleaseEvent(QGraphicsSceneMouseEvent* ev)
{
    if (!m_actions.isEmpty()) {
        foreach (const ActionInfo& action, m_actions) {
            if (action.actRect.contains(ev->pos())) {
                emit actionInvoked(m_id, action.actId);
                break;
            }
        }
    }
}

void NotifyTextItem::hoverMoveEvent(QGraphicsSceneHoverEvent* ev)
{
    uint hoveredId = 0;
    foreach (const ActionInfo& action, m_actions) {
        if (action.actRect.contains(ev->pos())) {
            setCursor(Qt::PointingHandCursor);
            hoveredId = action.actId;
            break;
        }
    }

    if (hoveredId != 0) {
        m_hoveredActId = hoveredId;
    } else {
        m_hoveredActId = 0;
        unsetCursor();
    }
    update();
}
