/***************************************************************************
 *   systemtraywidget.h                                                    *
 *                                                                         *
 *   Copyright (C) 2007 Jason Stubbs <jasonbstubbs@gmail.com>              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

// Own
#include "systemtraycontainer.h"

// KDE
#include <KDebug>
#include <plasma/theme.h>

// Qt
#include <QTimer>
#include <QX11Info>

// Xlib
#include <X11/Xlib.h>
#include <X11/Xutil.h>

SystemTrayContainer::SystemTrayContainer(QWidget *parent)
    : QX11EmbedContainer(parent)
    , m_updateTimer(0)
{
    connect(this, SIGNAL(clientClosed()), SLOT(deleteLater()));
    connect(this, SIGNAL(error(QX11EmbedContainer::Error)), SLOT(handleError(QX11EmbedContainer::Error)));

    connect(Plasma::Theme::defaultTheme(), SIGNAL(themeChanged()), this, SLOT(updateBackground()));
    updateBackground();

    // Tray icons have a fixed size of 22x22
    setMaximumSize(22, 22);
}

void SystemTrayContainer::embedSystemTrayClient( WId clientId )
{
    kDebug() << "attempting to embed" << clientId;
    if( !prepareFor(clientId)) { // temporary hack, until QX11EmbedContainer gets fixed
         deleteLater();
         return;
    }

    embedClient(clientId);

    // check if we still have a valid clientId since there may cases where we don't any
    // longer after calling embedClient like e.g. if there is already a pidgin-instance
    // running and it got started again. In that case those guniqueapplication starts
    // and fires a SYSTEM_TRAY_REQUEST_DOCK with another clientId up, exists and passes
    // commandline-arguments on to the other running instance and embedClient does fail
    // without emitting a clientClosed() or error() signal.
    XWindowAttributes attr;
    if( !XGetWindowAttributes(QX11Info::display(), clientId, &attr) /*|| attr.map_state == IsUnmapped*/ ) {
        deleteLater();
    }
}

bool SystemTrayContainer::x11Event(XEvent *event)
{
    bool ok = QX11EmbedContainer::x11Event(event);
    if (event->type == ReparentNotify) {
        setMinimumSize(22,22);
    }
    return ok;
}

void SystemTrayContainer::updateBackground()
{
    // Qt's regular quasi-transparent background doesn't work so set it to the
    // theme's background color instead.
    QPalette p = palette();
    p.setBrush(QPalette::Window, Plasma::Theme::defaultTheme()->color(Plasma::Theme::BackgroundColor));
    setPalette(p);
    setBackgroundRole(QPalette::Window);
}

void SystemTrayContainer::handleError(QX11EmbedContainer::Error error)
{
    Q_UNUSED(error);
    deleteLater();
}

// Temporary hack to change X window used by QX11EmbedContainer so that it matches
// the window embedded into it (#153193).
bool SystemTrayContainer::prepareFor(WId w)
{
    Display* dpy = QX11Info::display();

    XWindowAttributes ga;
    if( !XGetWindowAttributes(dpy, w, &ga))
        return false;

    XSetWindowAttributes sa;
    sa.background_pixel = WhitePixel(dpy, DefaultScreen(dpy));
    sa.border_pixel = BlackPixel(dpy, DefaultScreen(dpy));
    sa.colormap = ga.colormap;

    Window ww = XCreateWindow(dpy, parentWidget() ? parentWidget()->winId() : DefaultRootWindow(dpy),
            0, 0, 22, 22, 0, ga.depth, InputOutput, ga.visual,
            CWBackPixel | CWBorderPixel | CWColormap, &sa);
    create(ww, true, true);

    // repeat everything from QX11EmbedContainer's ctor that might be relevant
    setFocusPolicy(Qt::StrongFocus);
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    setAcceptDrops(true);
    setEnabled(false);

    XSelectInput(dpy, ww,
            KeyPressMask | KeyReleaseMask |
            ButtonPressMask | ButtonReleaseMask | ButtonMotionMask |
            KeymapStateMask |
            PointerMotionMask |
            EnterWindowMask | LeaveWindowMask |
            FocusChangeMask |
            ExposureMask |
            StructureNotifyMask |
            SubstructureNotifyMask);
    XFlush(dpy);
    return true;
}

void SystemTrayContainer::paintEvent(QPaintEvent *event)
{
    Q_UNUSED(event);

    if (!m_updateTimer) {
        m_updateTimer = new QTimer();
        connect(m_updateTimer, SIGNAL(timeout()),
                this, SLOT(delayedUpdate()));

        m_updateTimer->setSingleShot(true);
        m_updateTimer->setInterval(500);
        m_updateTimer->start();

        QTimer::singleShot(0, this, SLOT(updateBackgroundPixmap()));

    } else if (m_updateTimer->isActive()) {
        m_updateTimer->start();
    }
}

void SystemTrayContainer::delayedUpdate()
{
    updateBackgroundPixmap();
    m_updateTimer->deleteLater();
    m_updateTimer = 0;
}

void SystemTrayContainer::updateBackgroundPixmap()
{
    if (!testAttribute(Qt::WA_NativeWindow)) {
        return;
    }

    Pixmap bg = createBackgroundPixmap(winId());
    if (bg) {
        Display *display = QX11Info::display();

        XSetWindowBackgroundPixmap(display, winId(), bg);
        XFreePixmap(display, bg);

        sendExposeToClient();
    }
}

Qt::HANDLE SystemTrayContainer::createBackgroundPixmap(WId win)
{
    Display *display = QX11Info::display();

    XWindowAttributes attr;
    XGetWindowAttributes(display, win, &attr);

    QImage srcImage;
    switch (attr.depth) {
    case 32:
        srcImage = QImage(size(), QImage::Format_ARGB32);
        break;
    case 24:
        srcImage = QImage(size(), QImage::Format_RGB32);
        break;
    case 16:
        srcImage = QImage(size(), QImage::Format_RGB16);
        break;
    default:
        return 0;
    }

    QWidget *topWidget = this;
    while (topWidget->parentWidget()) {
        topWidget = topWidget->parentWidget();
    }

    topWidget->render(&srcImage, QPoint(0, 0), QRect(mapTo(topWidget, QPoint(0, 0)), size()));

    char *data = (char*)malloc(srcImage.numBytes());
    memcpy(data, srcImage.bits(), srcImage.numBytes());

    XImage *bgImage = XCreateImage(display, attr.visual, attr.depth, ZPixmap, 0,
            data, srcImage.width(), srcImage.height(), srcImage.depth(), srcImage.bytesPerLine());

    Pixmap bg = XCreatePixmap(display, win, srcImage.width(), srcImage.height(), attr.depth);

    GC gc = XCreateGC(display, bg, 0, 0);
    XPutImage(display, bg, gc, bgImage, 0, 0, 0, 0, srcImage.width(), srcImage.height());
    XFreeGC(display, gc);

    // XDestroyImage() frees the malloc'd data above too
    XDestroyImage(bgImage);

    return bg;
}

void SystemTrayContainer::sendExposeToClient()
{
    Display *display = QX11Info::display();

    XExposeEvent expose;
    expose.type = Expose;
    expose.serial = 0;
    expose.send_event = True;
    expose.display = display;
    expose.window = clientWinId();
    expose.x = 0;
    expose.y = 0;
    expose.width = width();
    expose.height = height();
    expose.count = 0;

    XSendEvent(display, clientWinId(), True, 0, (XEvent*)&expose);
}

#include "systemtraycontainer.moc"
